# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
# Copyright 2005 Dan Williams <dcbw@redhat.com> and Red Hat, Inc.

import time
import string
import xmlrpclib
import sys
import socket
import os
import threading
import urllib
import OpenSSL
from plague import FileDownloader
from plague import CommonErrors

DL_RETRIES = 'retries'
DL_STATUS = 'status'
DL_WAIT_TIME = 'wait_time'

STATUS_WAITING = 'waiting'
STATUS_ERROR = 'error'
STATUS_INPROGRESS = 'in-progress'
STATUS_DONE = 'done'

class ArchJob:
    """ Tracks a single build instance for a single arch on a builder """

    def __init__(self, builder, cfg, server, par_job, jobid, target_dict):
        self.par_job = par_job
        self.builder = builder
        self._server = server
        self._use_ssl = cfg.get_bool("Builders", "use_ssl")
        self.jobid = jobid
        self._status = 'running'
        self._builder_status = ''
        self._failure_noticed = False
        self._download_failed = False
        self._internal_failure = False
        self._target_dict = target_dict
        self._builder_gone = False
        self._download_lock = threading.Lock()
        self._downloads = {}
        self._starttime = time.time()
        self._endtime = 0
        self._die = False
        self._die_user_requested = False
        self._die_lock = threading.Lock()

        # SSL certificate and key filenames
        if self._use_ssl:
            self._certs = {}
            self._certs['key_and_cert'] = cfg.get_str("SSL", "server_key_and_cert")
            self._certs['ca_cert'] = cfg.get_str("SSL", "ca_cert")
            self._certs['peer_ca_cert'] = cfg.get_str("SSL", "ca_cert")
        else:
            self._certs = None

    def failure_noticed(self):
        return self._failure_noticed

    def set_failure_noticed(self):
        self._failure_noticed = True

    def _builder_finished(self):
        if self._builder_status == 'done' or self._builder_status == 'killed' or self._builder_status == 'failed' or self._builder_status == 'orphaned':
            return True
        return False

    def builder_failed(self):
        if self._builder_status == 'killed' or self._builder_status == 'failed':
            return True
        return False

    def download_failed(self):
        return self._download_failed

    def internal_failure(self):
        return self._internal_failure

    def builder_prepping(self):
        if self._builder_status == 'prepping':
            return True
        return False

    def arch(self):
        return self._target_dict['arch']

    def _to_dict(self):
        attrdict = {}
        attrdict['jobid'] = self.jobid
        attrdict['parent_uid'] = self.par_job.uid
        attrdict['arch'] = self._target_dict['arch']
        addr = self.builder.address()
        # for some reason, splithost doesn't like the protocol
        # method, you have to give it a string starting with "//"
        if addr.startswith("http"):
            idx = addr.find('//')
            addr = addr[idx:]
        host_port, path = urllib.splithost(addr)
        host, port = urllib.splitport(host_port)
        attrdict['builder_addr'] = host
        attrdict['status'] = self._status
        attrdict['builder_status'] = self._builder_status
        attrdict['starttime'] = self._starttime
        attrdict['endtime'] = self._endtime
        return attrdict

    def set_builder_job_status(self, builder_status):
        oldstatus = self._builder_status
        self._builder_status = builder_status
        if oldstatus != self._builder_status:
            attrdict = self._to_dict()
            self.par_job.bm.queue_archjob_status_update(self.jobid, attrdict)
            del attrdict

        if builder_status == 'killed' or builder_status == 'failed':
            self.par_job.wake()

    def _set_status(self, status):
        oldstatus = self._status
        self._status = status
        if oldstatus != self._status:
            attrdict = self._to_dict()
            self.par_job.bm.queue_archjob_status_update(self.jobid, attrdict)
            del attrdict

    def _send_repo_unlocked(self):
        success = False
        try:
            self._server.repo_unlocked(self.jobid)
            success = True
        except socket.error, e:
            if not CommonErrors.canIgnoreSocketError(e):
                print "%s (%s/%s): [ %s ] Unknown error sending repo unlocked: '%s'" % (self.par_job.uid,
                            self.par_job.package, self._target_dict['arch'], self.builder.address(), e)
        except socket.timeout, e:
            print "%s (%s/%s): [ %s ] Timeout sending repo unlocked: '%s'.  Trying again." % (self.par_job.uid,
                        self.par_job.package, self._target_dict['arch'], self.builder.address(), e)
        except xmlrpclib.ProtocolError, e:
            pass
        return success

    def _dl_files(self):
        files = []
        success = False
        try:
            files = self._server.files(self.jobid)
            success = True
        except socket.error, e:
            if not CommonErrors.canIgnoreSocketError(e):
                print "%s (%s/%s): [ %s ] Unknown error getting file list: '%s'" % (self.par_job.uid,
                            self.par_job.package, self._target_dict['arch'], self.builder.address(), e)
        except socket.timeout, e:
            print "%s (%s/%s): [ %s ] Timeout getting file list: '%s'" % (self.par_job.uid,
                        self.par_job.package, self._target_dict['arch'], self.builder.address(), e)
        except xmlrpclib.ProtocolError, e:
            pass
        return (success, files)

    def _is_done_status(self):
        if self._status == 'done':
            return True
        return False

    def _status_running(self):
        # Builders pause before they enter the 'prep' state (which accesses
        # the repo for this target), and wait for the server to allow them
        # to proceed when the repo is unlocked.
        if self._builder_status == 'downloaded':
            if not self.par_job.repo.locked():
                self._send_repo_unlocked()

        # if the builder is done, grab list of files to download
        if self._builder_finished():
            (success, files) = self._dl_files()
            if success:
                self._set_status('downloading')
                for f in files:
                    uf = urllib.unquote(f)
                    dl_dict = {}
                    dl_dict[DL_STATUS] = STATUS_WAITING
                    dl_dict[DL_RETRIES] = 0
                    dl_dict[DL_WAIT_TIME] = 0
                    self._downloads[uf] = dl_dict

    def dl_callback(self, status, cb_data, err_msg):
        url = cb_data
        self.builder.fwd_dl_callback(self)
        self._download_lock.acquire()
        dl_dict = self._downloads[url]
        if status == 'done':
            dl_dict[DL_STATUS] = STATUS_DONE
        elif status == 'failed':
            # Retry the download up to 10 times, then fail it
            if dl_dict[DL_RETRIES] >= 10:
                dl_dict[DL_STATUS] = STATUS_ERROR
            else:
                print "%s (%s/%s): Failed to retrieve %s (attempt %d) (Error %s), trying again..." % (self.par_job.uid,
                        self.par_job.package, self._target_dict['arch'], url, dl_dict[DL_RETRIES], err_msg)
                dl_dict[DL_STATUS] = STATUS_WAITING
                dl_dict[DL_WAIT_TIME] = 5  # Wait a bit before trying again
                dl_dict[DL_RETRIES] = dl_dict[DL_RETRIES] + 1
        self._download_lock.release()

    def _print_downloaded_files(self):
        file_string = ""
        ndownloads = len(self._downloads.keys())
        for url in self._downloads.keys():
            filename = os.path.basename(url)
            string = "'" + filename + "'"
            dl_dict = self._downloads[url]
            if dl_dict[DL_STATUS] == STATUS_ERROR:
                string = string + " (failed)"
            file_string = file_string + string
            if url != self._downloads.keys()[ndownloads - 1]:
                file_string = file_string + ", "

        print "%s (%s/%s): Build result files - [ %s ]" % (self.par_job.uid,
                    self.par_job.package, self._target_dict['arch'], file_string)

    def _status_downloading(self):
        # Start grabbing the next undownloaded file, but only
        # if we aren't already pulling one down
        undownloaded = False
        failed = False
        self._download_lock.acquire()
        for url in self._downloads.keys():
            dl_dict = self._downloads[url]
            dl_status = dl_dict[DL_STATUS]
            if dl_status == STATUS_WAITING:
                # If the download got retried due to a previous
                # download error, we may have to wait a bit
                if dl_dict[DL_WAIT_TIME] > 0:
                    dl_dict[DL_WAIT_TIME] = dl_dict[DL_WAIT_TIME] - 1
                    undownloaded = True
                    continue

                # Otherwise, spawn the download thread to grab the file
                target_dir = os.path.join(self.par_job.get_stage_dir(), self._target_dict['arch'])
                if not os.path.exists(target_dir):
                    os.makedirs(target_dir)

                try:
                    dl_thread = FileDownloader.FileDownloader(self.dl_callback, url, url,
                                    target_dir, ['.rpm', '.log'], self._certs)
                    dl_thread.start()
                    undownloaded = True
                    dl_dict[DL_STATUS] = STATUS_INPROGRESS
                except FileDownloader.FileNameException, e:
                    print "%s (%s/%s): [ %s ] Bad file name error getting %s: '%s'" % (self.par_job.uid,
                                self.par_job.package, self._target_dict['arch'], self.builder.address(), url, e)
                    # Hard error, we don't retry this one
                    dl_dict[DL_STATUS] = STATUS_ERROR
                break
            elif dl_status == STATUS_INPROGRESS:
                undownloaded = True
                break
            elif dl_status == STATUS_ERROR:
                failed = True
                continue
            elif dl_status == STATUS_DONE:
                continue
        self._download_lock.release()

        # All done downloading?
        if not undownloaded:
            self._print_downloaded_files()
            self._endtime = time.time()
            if failed:
                self._download_failed = True
            self._set_status('done')
            self.par_job.wake()

    def process(self):
        if self._is_done_status():
            return

        # If we're supposed to die, tell the builder and clean up
        self._die_lock.acquire()
        should_die = self._die
        user_requested = self._die_user_requested
        self._die_lock.release()
        if should_die:
            try:
                self._server.die(self.jobid)
            except (socket.error, socket.timeout, OpenSSL.SSL.SysCallError, OpenSSL.SSL.Error):
                pass
            self._set_status('done')
            if user_requested:
                print "%s (%s/%s): %s - killed." % (self.par_job.uid, self.par_job.package,
                        self._target_dict['arch'], self.jobid)
            return

        status_func = None
        try:
            status_func = getattr(self, "_status_%s" % self._status)
        except AttributeError:
            print "%s (%s/%s): %s - internal archjob inconsistency.  Unknown status '%s'." % (self.par_job.uid, self.par_job.package,
                        self._target_dict['arch'], self.jobid, self._status)
            self._set_status('done')
            self._internal_failure = True
            return
            
        # Do the actual work for this status
        status_func()

    def get_status(self):
        return self._status

    def get_files(self):
        """ Return a list of base filenames we got from the builder """
        files = []
        for url in self._downloads.keys():
            try:
                fname = FileDownloader.get_base_filename_from_url(url, ['.rpm', '.log'])
                dl_dict = self._downloads[url]
                dl_status = dl_dict[DL_STATUS]
                if dl_status == STATUS_DONE:
                    files.append(fname)
            except FileDownloader.FileNameException, e:
                # Just ignore the file then
                print "%s (%s/%s): Illegal file name.  Error: '%s', URL: %s" % (self.par_job.uid,
                        self.par_job.package, self._target_dict['arch'], e, url)
                pass
        return files

    def builder_gone(self):
        if self._status != 'done':
            self._builder_status = 'orphaned'
            self._set_status('done')
            self.par_job.remove_arch_job(self)

    def die(self, user_requested=False):
        # Can be called from other threads
        if self._status == 'running':
            self._die_lock.acquire()
            self._die = True
            self._die_user_requested = user_requested
            self._die_lock.release()
            if user_requested:
                print "%s (%s/%s): %s - kill requested by parent job" % (self.par_job.uid,
                        self.par_job.package, self._target_dict['arch'], self.jobid)


