# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
# Copyright 2005 Dan Williams <dcbw@redhat.com> and Red Hat, Inc.

import time
import string
import xmlrpclib
import sys
import socket
import os
import urllib
import threading
from plague import XMLRPCServerProxy
from plague import CommonErrors
import OpenSSL
import ArchJob
import EmailUtils
import Config
from plague import DebugUtils

SUSPEND_NONE = 'none'
SUSPEND_TIMEOUT = 'timeout'
SUSPEND_HARD_ERROR = 'hard-error'

# HACK: This class is a hack to work around SSL hanging issues,
# which cause the whole server to grind to a halt
class BuildingJobsCheck(threading.Thread):
    def __init__(self, server, address):
        self._server = server
        self._address = address

        self.done = False
        self.failed = False

        threading.Thread.__init__(self)
        self.setName("BuildingJobsCheck: %s" % address)

    def run(self):
        jobs = {}
        free_slots = 0
        try:
            (jobs, free_slots) = self._server.building_jobs()
        except (socket.error, socket.timeout, OpenSSL.SSL.SysCallError, OpenSSL.SSL.Error, xmlrpclib.ProtocolError):
            self.failed = True
        except xmlrpclib.Fault, e:
            print "Builder Error (%s) in _building_jobs(): builder replied '%s'" % (self._address, e)
            self.failed = True
        self.jobs = jobs
        self.free_slots = free_slots
        self.done = True


class Builder(threading.Thread):
    """ Tracks all jobs on a builder instance """

    _BUILDER_PING_INTERVAL = 60 * 5      # In seconds

    def __init__(self, manager, cfg, address):
        self._cur_jobid = None
        self._manager = manager
        self._jobs = {}
        self._free_slots = 0
        self._num_slots = 0
        self._address = address
        self._alive = True
        self._suspend_reason = SUSPEND_NONE
        self._stop = False
        self._prepping_jobs = False
        self._unavail_count = 0
        self._target_list = []
        self._ping_timeout = 0
        self._cur_ping_interval = self._BUILDER_PING_INTERVAL
        self._ping_now = False
        self._when_died = 0
        self._server_cfg = cfg

        self._archjob_dl_lock = threading.Lock()
        self._archjob_dl_queue = []

        certs = {}
        if self._server_cfg.get_bool("Builders", "use_ssl"):
            certs['key_and_cert'] = self._server_cfg.get_str("SSL", "server_key_and_cert")
            certs['ca_cert'] = self._server_cfg.get_str("SSL", "ca_cert")
            certs['peer_ca_cert'] = self._server_cfg.get_str("SSL", "ca_cert")
        else:
            certs = None

        self._server = XMLRPCServerProxy.PlgXMLRPCServerProxy(self._address, certs, timeout=20)
        self._server_lock = threading.Lock()

        threading.Thread.__init__(self)
        self.setName("Builder: %s" % address)

        (self._alive, target_list) = self._ping_builder()
        if self._alive:
            self._init_builder(target_list)
        else:
            # Treat the builder as timed out and ping it periodically
            self._ping_timeout = time.time()
            self._suspend_reason = SUSPEND_TIMEOUT

    def _init_builder(self, target_list):
        self._target_list = target_list

        # Kill any jobs currently running on the builder
        jobs = self._building_jobs()
        for jobid in jobs.keys():
            try:
                self._server.die(jobid)
            except:
                pass

        self._num_slots = self._get_num_slots()

    def _get_num_slots(self):
        num_slots = self._num_slots
        try:
            num_slots = self._server.num_slots()
        except (socket.error, socket.timeout, OpenSSL.SSL.SysCallError, OpenSSL.SSL.Error, xmlrpclib.ProtocolError):
            pass
        except xmlrpclib.Fault, e:
            print "Builder Error (%s) in _get_num_slots(): builder replied '%s'" % (self._address, e)
        return num_slots

    def _building_jobs(self):
        bjc = BuildingJobsCheck(self._server, self._address)

        curtime = time.time()
        bjc.start()

        # Give the check 10s, otherwise screw it
        while time.time() - curtime < 10:
            if bjc.done:
                break
            time.sleep(0.5)

        if bjc.done:
            if not bjc.failed:
                self._unavail_count = 0
                self._alive = True
                self._free_slots = bjc.free_slots
                return bjc.jobs
            else:
                # Error of some kind
                self._unavail_count = self._unavail_count + 1

        return {}

    def _ping_builder(self):
        target_list = []
        try:
            target_list = self._server.supported_targets()
            alive = True
        except (socket.error, socket.timeout, OpenSSL.SSL.SysCallError, OpenSSL.SSL.Error, xmlrpclib.ProtocolError):
            alive = False
        except xmlrpclib.Fault, e:
            print "Builder Error (%s) in _ping_builder(): builder replied '%s'" % (self._address, e)
            alive = False
        return (alive, target_list)

    def _match_target_dict(self, td1, td2):
        if td1['distro'] == td2['distro']:
            if td1['target'] == td2['target']:
                if td1['repo'] == td2['repo']:
                    return True
        return False

    def arches(self, target_dict):
        for td in self._target_list:
            if self._match_target_dict(td, target_dict):
                arches = []
                for arch in td['supported_arches']:
                    if not arch in arches:
                        arches.append(arch)
                return arches
        return None

    def can_build_for_target(self, target_dict):
        for td in self._target_list:
            if self._match_target_dict(td, target_dict):
                if target_dict['arch'] in td['supported_arches']:
                    return True
        return False

    def address(self):
        return self._address

    def alive(self):
        """ Is the builder responding to requests? """
        return self._alive

    def free_slots(self):
        return self._free_slots
    
    def start_job(self, par_job, target_dict, srpm_url):
        if not self.available():
            raise RuntimeError
        if not self.can_build_for_target(target_dict):
            raise RuntimeError

        self._server_lock.acquire()
        try:
            # Builder will return jobid of 0 if it can't start the job for some reason
            jobid = self._server.start_new_job(target_dict, srpm_url)
        except (socket.error, socket.timeout, OpenSSL.SSL.SysCallError, \
                OpenSSL.SSL.Error, xmlrpclib.ProtocolError, xmlrpclib.Fault), e:
            error_class = str(e.__class__)
            error_string = str(e)
            jobarch = target_dict['arch']
            print "%s (%s/%s): %s exception '%s' starting job on %s" % (par_job.uid, \
                par_job.package, jobarch, error_class, error_string, \
                self._address)
            # Check for hard errors, for which we suspend the builder
            if string.find(error_string, "OSError") >= 0 and string.find(error_string, "Errno") >= 0:
                self._handle_builder_suspend(SUSPEND_HARD_ERROR, error_string)
            time.sleep(0.5)
            jobid = 0

        if jobid == 0:
            self._server_lock.release()
            raise RuntimeError

        job = ArchJob.ArchJob(self, self._server_cfg, self._server, par_job, jobid, target_dict)
        self._jobs[jobid] = job
        self._update_building_jobs()
        self._server_lock.release()
        return job

    def _update_building_jobs(self):
        jobs = self._building_jobs()

        # Update the current job's status
        if self._unavail_count == 0:
            self._prepping_jobs = False
            builder_jobs = []
            for jobid in jobs.keys():
                try:
                    job = self._jobs[jobid]
                    status = jobs[jobid]
                    job.set_builder_job_status(status)
                    if status == 'prepping':
                        self._prepping_jobs = True
                    builder_jobs.append(jobid)
                except KeyError:
                    pass

            # We have to check jobs that weren't reported
            # as 'building' by the builder, since the job
            # may have finished on the builder and was
            # removed from the building job list before we
            # were able to know that it was done.  HACK
            for jobid in self._jobs.keys():
                # If the builder didn't report this job as building,
                # and its not done, explicitly grab its status
                job = self._jobs[jobid]
                if jobid not in builder_jobs and job.get_status() != 'done':
                    status = self._get_builder_job_status(jobid)
                    if status:
                        job.set_builder_job_status(status)

    def _get_builder_job_status(self, jobid):
        """ Get the status of one job on the builder """
        status = None
        try:
            status = self._server.job_status(jobid)
        except (socket.error, socket.timeout, OpenSSL.SSL.SysCallError, OpenSSL.SSL.Error):
            pass
        except (xmlrpclib.Fault, xmlrpclib.ProtocolError):
            pass
        return status

    def stop(self):
        self._stop = True

    def ping_asap(self):
        # Reduce the ping interval to ping the builder right away
        self._cur_ping_interval = 0
        self._ping_now = True

    def _handle_builder_suspend(self, reason, msg):
        for jobid in self._jobs.keys():
            job = self._jobs[jobid]
            job.builder_gone()
            del self._jobs[jobid]
        self._jobs = {}
        self._alive = False
        self._suspend_reason = reason
        self._unavail_count = 0
        self._prepping_jobs = False
        self._when_died = time.time()

        # Reset current ping interval to default
        self._cur_ping_interval = self._BUILDER_PING_INTERVAL
        self._ping_timeout = time.time()

        # Notify admins
        print "Suspending builder '%s'.  Reason: %s - %s." % (self._address, reason, msg)
        subject = "Builder Suspended: %s" % self._address
        msg = "The builder '%s' was suspended.  Reason: %s - %s." % (self._address, reason, msg)
        sender = self._server_cfg.get_str("Email", "email_from")
        for addr in self._server_cfg.get_list("Email", "admin_emails"):
            EmailUtils.email_result(sender, addr, msg, subject)

    def _handle_builder_reactivate(self, target_list):
        self._alive = True
        self._suspend_reason = SUSPEND_NONE
        self._ping_timeout = 0

        self._init_builder(target_list)

        print "Re-activating builder '%s'." % self._address
        subject = "Builder Re-activated: %s" % self._address
        msg = """The builder '%s' was re-activated.

  Suspended at: %s
  Re-Enabled at: %s
""" % (self._address, time.ctime(self._when_died), time.ctime(time.time()))
        sender = self._server_cfg.get_str("Email", "email_from")
        for addr in self._server_cfg.get_list("Email", "admin_emails"):
            EmailUtils.email_result(sender, addr, msg, subject)
        self._when_died = 0

    def fwd_dl_callback(self,job):
        self._archjob_dl_lock.acquire()
        self._archjob_dl_queue.append(job)
        self._archjob_dl_lock.release()
    
    def run(self):
        DebugUtils.registerThreadName(self)
        while not self._stop:
            self._server_lock.acquire()

            if self._alive:
                self._update_building_jobs()

                if self._unavail_count > 2:
                    # Kill all jobs on the client if it went away
                    self._handle_builder_suspend(SUSPEND_TIMEOUT, "the builder timed out")
                else:
                    # Update status of all archjobs on this builder
                    for j in self._jobs.values():
                        j.process()
            elif not self._alive and (self._suspend_reason == SUSPEND_TIMEOUT or self._ping_now):
                # Ping the builder every so often to see if it responds again
                if time.time() > (self._ping_timeout + self._cur_ping_interval):
                    (alive, target_list) = self._ping_builder()
                    if alive:
                        self._handle_builder_reactivate(target_list)
                    else:
                        # Wait and ping again
                        self._ping_timeout = time.time()

                    # Reset current ping interval to default
                    self._cur_ping_interval = self._BUILDER_PING_INTERVAL
                    self._ping_now = False

            self._server_lock.release()

            last_time = time.time()
            while (time.time() <= last_time+20):

                # Process archjobs which had download callbacks,
                # so next file download is started sooner.
                self._archjob_dl_lock.acquire()
                dl_jobs = self._archjob_dl_queue
                self._archjob_dl_queue = []
                self._archjob_dl_lock.release()
                if len(dl_jobs):
                    status_change = False
                    self._server_lock.acquire()
                    for j in dl_jobs:
                        old_status = j.get_status()
                        j.process()
                        if old_status != j.get_status():
                            status_change = True
                    self._server_lock.release()
                    if status_change:
                        break

                time.sleep(1)
    
    def available(self):
        """
        Can the builder start a new job right now?
        """
        if self._unavail_count > 2 or not self._alive or self.free_slots() <= 0:
            return False
        return True

    def any_prepping_jobs(self):
        return self._prepping_jobs

    def to_dict(self):
        builder_dict = {}

        addr = self._address
        # for some reason, splithost doesn't like the protocol
        # method, you have to give it a string starting with "//"
        if addr.startswith("http"):
            idx = addr.find('//')
            addr = addr[idx:]
        host_port, path = urllib.splithost(addr)
        host, port = urllib.splitport(host_port)
        builder_dict['address'] = host

        arches = []
        for td in self._target_list:
            for arch in td['supported_arches']:
                if not arch in arches:
                    arches.append(arch)
        builder_dict['arches'] = arches

        builder_dict['alive'] = self._alive
        builder_dict['num_slots'] = self._num_slots
        builder_dict['free_slots'] = self._free_slots
        return builder_dict

