# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
# Copyright 2005 Dan Williams <dcbw@redhat.com> and Red Hat, Inc.

import time
import string
import xmlrpclib
import sys
import socket
import os
import threading
import Builder
import EmailUtils
import Config


class BuilderManager:
    """
    Tracks individual builder instances.
    """

    def __init__(self, cfg):
        self._cfg = cfg
        self._builders_lock = threading.Lock()

        self._builders = []
        self.add_new_builders()

        # Print out builder list when starting up
        print "\nBuilders:"
        print "-" * 90
        for builder in self._builders:
            string = "  " + builder.address()
            string = string + " " * (40 - len(builder.address()))
            builder_dict = builder.to_dict()
            for arch in builder_dict['arches']:
                string = string + arch + " "
            string = string + " " * (80 - len(string))
            status = "unavailable"
            if builder_dict['alive']:
                status = "alive"
            string = string + status
            del builder_dict
            print string
        print ""

        self._queue_lock = threading.Lock()
        self._queue = []

        self._have_work = False

    def __del__(self):
        for builder in self._builders:
            builder.stop()
        time.sleep(2)

    def set_build_master(self, build_master):
        self._build_master = build_master

    def add_new_builders(self):
        self._builders_lock.acquire()

        builder_list = []
        tmp_list = self._cfg.get_list("Builders", "builders")
        prefix = "http://"
        if self._cfg.get_bool("Builders", "use_ssl") == True:
            prefix = "https://"
        for addr in tmp_list:
            # Rewrite addresses to match current builder connection method
            if addr.startswith("http://"):
                addr = addr[7:]
            elif addr.startswith("https://"):
                addr = addr[8:]

            builder_list.append(prefix + addr)

        for address in builder_list:
            # If the address is already in our _builders list, skip it
            skip = False
            for builder in self._builders:
                if address == builder.address():
                    skip = True
            if skip == True:
                continue

            # Add the builder to our build list
            builder = Builder.Builder(self, self._cfg, address)
            builder.start()
            self._builders.append(builder)

        self._builders_lock.release()

    def ping_suspended_builders(self):
        self._builders_lock.acquire()
        for builder in self._builders:
            if not builder.alive():
                builder.ping_asap()
        self._builders_lock.release()

    def list_builders(self):
        builder_list = []
        for builder in self._builders:
            builder_list.append(builder.to_dict())
        return builder_list

    def have_work(self, paused):
        avail = False
        for builder in self._builders:
            if builder.available():
                avail = True
        if not paused and len(self._queue) > 0 and avail:
            return True
        return self._have_work

    def _builder_sort_func(self, builder):
        return builder.free_slots()

    def _builder_cmp_func(self, builder1, builder2):
        if builder1.free_slots() > builder2.free_slots():
            return 1
        elif builder1.free_slots() == builder2.free_slots():
            return 0
        elif builder1.free_slots() < builder2.free_slots():
            return -1
        return 1

    def process(self, paused):
        self._have_work = False

        # Don't queue any new jobs if we are paused
        if paused:
            return

        # Deal with new arch jobs
        self._queue_lock.acquire()
        new_jobs = {}
        for req in self._queue:
            parent = req['parent']
            stage = parent.get_cur_stage()
            if (stage != 'prep') and (stage != 'building') and (stage != 'waiting'):
                self._queue.remove(req)
                continue

            # Find all free builders that could satisfy the request
            builder_list = []
            for builder in self._builders:
                if builder.available() and builder.can_build_for_target(req['target_dict']):
                    builder_list.append(builder)

            # Use the builder with the most free build slots
            if sys.version_info[:3] >= (2, 4, 0):
                builder_list.sort(key=self._builder_sort_func, reverse=True)
            else:
                builder_list.sort(self._builder_cmp_func)
                builder_list.reverse()

            for builder in builder_list:
                try:
                    job = builder.start_job(parent, req['target_dict'], req['srpm_url'])
                except RuntimeError:
                    pass
                else:
                    if not new_jobs.has_key(parent):
                        new_jobs[parent] = []
                    new_jobs[parent].append(job)
                    self._queue.remove(req)
                    break
        self._queue_lock.release()

        # Notify the parent jobs of their new archjobs.  Have to do this outside _queue_lock
        # for locking reasons
        for parent in new_jobs.keys():
            for job in new_jobs[parent]:
                parent.add_arch_job(job)

        if len(self._queue) > 0:
            time.sleep(0.25)

    def request_arch_job(self, par_job, target_dict, srpm_url, orphaned):
        req = {}
        req['parent'] = par_job
        req['target_dict'] = target_dict
        req['srpm_url'] = srpm_url

        self._queue_lock.acquire()
        if orphaned:
            # insert orphaned requests at the front of the queue
            self._queue.insert(0, req)
        else:
            self._queue.append(req)
        self._queue_lock.release()

    def any_prepping_builders(self):
        # query each build builder for any jobs that are in the 'prepping' state
        for builder in self._builders:
            if builder.alive() and builder.any_prepping_jobs():
                return True
        return False


