# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
# Copyright 2005 Dan Williams <dcbw@redhat.com> and Red Hat, Inc.

import os
import threading
import shutil
import time
import commands
import popen2
import fcntl
import stat
import EmailUtils
from plague import DebugUtils

# Lockfile used by external scripts to ensure mutual exclusion
# from concurrent access to the repository's directory
REPO_LOCKFILE_NAME = ".repo-update.lock"

class Repo(threading.Thread):
    """ Represents an on-disk repository of RPMs and manages updates to the repo. """

    def __init__(self, target_cfg, repodir, builder_manager):
        self._builder_manager = builder_manager
        self._target_cfg = target_cfg
        self._target_dict = target_cfg.target_dict()
        if not os.path.exists(repodir):
            print "Error: Repository directory '%s' does not exist." % repodir
            os._exit(1)

        target_str = self._target_cfg.target_string()
        self._repodir = os.path.join(repodir, target_str)
        if not os.path.exists(self._repodir):
            os.makedirs(self._repodir)
        self._lockfile_path = os.path.join(self._repodir, REPO_LOCKFILE_NAME)

        self._repo_cache_dir = os.path.join(repodir, "cache", target_str)
        if not os.path.exists(self._repo_cache_dir):
            os.makedirs(self._repo_cache_dir)

        parent_cfg = self._target_cfg.parent_cfg()
        upload_dir = os.path.join(parent_cfg.get_str("Directories", "server_work_dir"), "srpm_upload_dir", target_str)
        if not os.path.exists(upload_dir):
            os.makedirs(upload_dir)
            os.chmod(upload_dir, stat.S_IRWXU | stat.S_IRWXG | stat.S_IRWXO)

        self._lock = threading.Lock()
        self._repo_additions = []
        self._lock_count = 0
        self._stop = False

        self._pobj = None
        self._repo_script_start = 0
        self._repo_script = None
        script = self._target_cfg.get_str("General", "repo_script")
        if len(script):
            self._repo_script = script

        threading.Thread.__init__(self)
        self.setName("Repo: %s" % target_str)

    def target_cfg(self):
        return self._target_cfg

    def request_copy(self, buildjob):
        """ Registers a BuildJob object that has files to copy to the repo """

        self._lock.acquire()
        self._repo_additions.append(buildjob)
        # We enter lock level 1 here, preventing builders from
        # starting their 'prep' state
        if self._lock_count == 0:
            self._lock_count = 1
        self._lock.release()

    def locked(self):
        # We can get away without holding _lock here...
        if self._lock_count > 0:
            return True
        return False

    def _update_repo_with_pushlock(self):
        lockfile = None
        try:
            lockfile = open(self._lockfile_path, 'w')
            rc = fcntl.flock(lockfile, fcntl.LOCK_EX)
        except IOError, (errno, strerr):
            target_str = self._target_cfg.target_string()
            print "Repo Error (%s): opening lockfile %s failed.  Output: (errno %d) '%s'" % (target_str, self._lockfile_path, errno, strerr)

        try:
            self._update_repo()
        except:
            if lockfile:
                fcntl.flock(lockfile, fcntl.LOCK_UN)
                lockfile.close()
            raise
        if lockfile:
            fcntl.flock(lockfile, fcntl.LOCK_UN)
            lockfile.close()

    def _update_repo(self):
        """ Copy new RPMS to each repo, and update each repo at the end """
        for buildjob in self._repo_additions:
            # Ensure all the files are accessible
            success = True
            bad_file = None
            for src in buildjob.repofiles.keys():
                if not os.path.exists(src) or not os.access(src, os.R_OK):
                    success = False
                    bad_file = src

            if success:
                for src in buildjob.repofiles.keys():
                    dst = buildjob.repofiles[src]
                    if not os.path.exists(os.path.dirname(dst)):
                        os.makedirs(os.path.dirname(dst))

                    file_in_dst = os.path.join(os.path.dirname(dst), os.path.basename(src))

                    if src.endswith(".src.rpm"):
                        # Only copy SRPMs to the repo dir if there's not already one there
                        if not os.path.exists(file_in_dst):
                            shutil.copy(src, file_in_dst)
                    else:
                        shutil.copy(src, file_in_dst)

            # Notify the build job that we've copied its files to the repo
            buildjob.repo_add_callback(success, bad_file)

        self._repo_additions = []

        (s, o) = commands.getstatusoutput('/usr/bin/createrepo -d -q -c %s -x "*.src.rpm" -x "*-debuginfo-*.rpm" %s' % (self._repo_cache_dir, self._repodir))
        if s != 0:
            target_str = self._target_cfg.target_string()
            print "Repo Error (%s): createrepo failed with exit status %d!  Output: '%s'" % (target_str, s, o)


    def _run_repo_script(self):
        target_str = self._target_cfg.target_string()
        cmd = "%s %s" % (self._repo_script, target_str)
        print "Repo '%s': executing repository script %s" % (target_str, self._repo_script)
        self._pobj = popen2.Popen4(cmd=cmd)
        fcntl.fcntl(self._pobj.fromchild.fileno(), fcntl.F_SETFL, os.O_NONBLOCK)
        self._repo_script_start = time.time()

    def _email_repo_script_failure(self, subject):
        server_cfg = self._target_cfg.parent_cfg()
        admins = server_cfg.get_list("Email", "admin_emails")
        sender = server_cfg.get_str("Email", "email_from")
        msg = self._get_repo_script_output()
        for addr in admins:
            EmailUtils.email_result(sender, addr, msg, subject)

    def _get_repo_script_output(self):
        output = ""
        while True:
            try:
                string = os.read(self._pobj.fromchild.fileno(), 1024)
                if not len(string):
                    break
            except OSError, e:
                break
            output = output + string
        return output

    def _monitor_repo_script(self):
        target_str = self._target_cfg.target_string()
        unlock = False
        exit_status = self._pobj.poll()
        if exit_status == 0:
            print "Repo '%s': repo script %s done." % (target_str, self._repo_script)
            unlock = True
        elif exit_status > 0:
            subj = "Repo '%s': repo script %s exited with error: %d." % (target_str, self._repo_script, exit_status)
            self._email_repo_script_failure(subj)
            print subj
            unlock = True
        else:
            # If the repo script has been going for more than an hour, kill it
            if time.time() > self._repo_script_start + (60 * 60):
                try:
                    os.kill(self._pobj.pid, 9)
                except OSError:
                    pass
                subj = "Repo '%s': repo script %s timed out and was killed." % (target_str, self._repo_script)
                self._email_repo_script_failure(subj)
                print subj
                unlock = True

        if unlock:
            self._repo_script_start = 0
            self._lock_count = 0
            self._pobj = None

    def run(self):
        DebugUtils.registerThreadName(self)
        while self._stop == False:
            # We have 3 lock levels:
            #
            # 0 - repo unlocked
            # 1 - entered when jobs request packages to be copied to the repo;
            #           builders blocked from entering the 'prep' state
            # 2 - entered when no builders are currently in the 'prep' state;
            #       packages copied to repo and createrepo is run
            # 3 - entered when createrepo is done; repo script run

            prepping_builders = self._builder_manager.any_prepping_builders()

            self._lock.acquire()

            # Level 2: update the repo
            if self._lock_count == 2:
                target_str = self._target_cfg.target_string()
                print "Repo '%s': updating repository metadata..." % target_str
                try:
                    self._update_repo_with_pushlock()
                except IOError, (err, strerr):
                    print "Repo '%s': %s (%d)" % (target_str, strerr, err)
                except OSError, e:
                    print "Repo '%s': %s" % (target_str, e)
                print "Repo '%s': Done updating." % target_str

                # If there's a repo script for this target, enter level 3
                if self._repo_script:
                    self._run_repo_script()
                    self._lock_count = 3
                else:
                    self._lock_count = 0

            # Level 3: monitor the repo script
            if self._lock_count == 3:
                if self._pobj:
                    self._monitor_repo_script()
                else:
                    # If for some reason self._pobj is None, unlock the repo
                    self._lock_count = 0

            # Enter lock level 2 if there are no builders in the
            # 'prep' state and we are already at lock level 1
            if not prepping_builders and self._lock_count == 1:
                self._lock_count = 2

            self._lock.release()

            time.sleep(5)

    def stop(self):
        self._stop = True

